##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
A lovely bunch of gates and instructions for programming with.  This module is used to provide
Pythonic sugar for Quil instructions.
"""

from .quilbase import (Measurement, Gate, Addr, Wait, Reset, Halt, Nop, ClassicalTrue,
                       ClassicalFalse, ClassicalNot, ClassicalAnd, ClassicalOr, ClassicalMove,
                       ClassicalExchange, unpack_qubit)
from six import integer_types


def unpack_classical_reg(c):
    """
    Get the address for a classical register.

    :param c: A list of length 1 or an int or an Addr.
    :return: The address as an Addr.
    """
    if not (isinstance(c, integer_types) or isinstance(c, (list, Addr))):
        raise TypeError("c should be an int or list or Addr")
    if isinstance(c, list) and (len(c) != 1 or not isinstance(c[0], int)):
        raise ValueError("if c is a list, it should be of 1 int")
    if isinstance(c, Addr):
        return c
    elif isinstance(c, list):
        return Addr(c[0])
    else:
        return Addr(c)


def _make_gate(name, num_qubits, num_params=0):
    def gate_function(*params):
        params = list(params)
        stray_qubits = []
        if len(params) < num_params:
            raise ValueError(
                "Wrong number of params for {}. {} given, require {}."
                .format(name, len(params), num_params)
            )
        elif len(params) > num_params:
            stray_qubits = params[num_params:]
            params = params[0:num_params]

        def ctor(*qubits):
            qubits = stray_qubits + list(qubits)
            if len(qubits) != num_qubits:
                raise ValueError(
                    "Wrong number of qubits for {}. {} given, require {}."
                    .format(name, len(qubits), num_qubits)
                )
            return Gate(name, params, [unpack_qubit(q) for q in qubits])

        if len(stray_qubits) == num_qubits:
            return Gate(name, params, [unpack_qubit(q) for q in stray_qubits])
        else:
            return ctor

    return gate_function


I = _make_gate("I", 1)()
"""
Produces the I instruction. This gate is a single qubit identity gate.
Note that this gate is different that the NOP instruction as noise channels
are typically still applied during the duration of identity gates. Identities will
also block parallelization like any other gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
X = _make_gate("X", 1)()
"""
Produces the X instruction. This gate is a single qubit X-gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
Y = _make_gate("Y", 1)()
"""
Produces the Y instruction. This gate is a single qubit Y-gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
Z = _make_gate("Z", 1)()
"""
Produces the Z instruction. This gate is a single qubit Z-gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
H = _make_gate("H", 1)()
"""
Produces the H instruction. This gate is a single qubit Hadamard gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
S = _make_gate("S", 1)()
"""
Produces the S instruction. This gate is a single qubit S-gate.

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
T = _make_gate("T", 1)()
"""
Produces the T instruction. This gate is a single qubit T-gate. It is the same
as RZ(pi/4).

:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""

RX = _make_gate("RX", 1, 1)
"""
Produces the RX instruction. This gate is a single qubit X-rotation.

:param angle: The angle to rotate around the x-axis on the bloch sphere.
:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
RY = _make_gate("RY", 1, 1)
"""
Produces the RY instruction. This gate is a single qubit Y-rotation.

:param angle: The angle to rotate around the y-axis on the bloch sphere.
:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
RZ = _make_gate("RZ", 1, 1)
"""
Produces the RZ instruction. This gate is a single qubit Z-rotation.

:param angle: The angle to rotate around the z-axis on the bloch sphere.
:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""
PHASE = _make_gate("PHASE", 1, 1)
"""
Produces a PHASE instruction. This is the same as the RZ gate.

:param angle: The angle to rotate around the z-axis on the bloch sphere.
:param qubit: The qubit apply the gate to.
:returns: A Gate object.
"""

CZ = _make_gate("CZ", 2)()
"""
Produces a CZ instruction.
This gate applies to two qubit arguments to produce the controlled-Z gate instruction.

:param control: The control qubit.
:param target: The target qubit. The target qubit has an Z-gate applied to it if the control qubit is in
               the excited state.
:returns: A Gate object.
"""

CNOT = _make_gate("CNOT", 2)()
"""
Produces a CNOT instruction.
This gate applies to two qubit arguments to produce the controlled-not gate instruction.

:param control: The control qubit.
:param target: The target qubit. The target qubit has an X-gate applied to it if the control qubit is in
               the excited state.
:returns: A Gate object.
"""
CCNOT = _make_gate("CCNOT", 3)()
"""
Produces a CCNOT instruction.
This gate applies to three qubit arguments to produce the controlled-controlled-not gate instruction.

:param control-1: The first control qubit.
:param control-2: The second control qubit.
:param target: The target qubit. The target qubit has an X-gate applied to it if both control qubits are in
               the excited state.
:returns: A Gate object.
"""

CPHASE00 = _make_gate("CPHASE00", 2, 1)
"""
Produces a CPHASE00 instruction.
This gate applies to two qubit arguments to produce one of the controlled phase instructions.

:param angle: The input phase angle to apply when both qubits are in the ground state.
:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
CPHASE01 = _make_gate("CPHASE01", 2, 1)
"""
Produces a CPHASE01 instruction.
This gate applies to two qubit arguments to produce one of the controlled phase instructions.

:param angle: The input phase angle to apply when q1 is in the excited state and q2 is in the ground state.
:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
CPHASE10 = _make_gate("CPHASE10", 2, 1)
"""
Produces a CPHASE10 instruction.
This gate applies to two qubit arguments to produce one of the controlled phase instructions.

:param angle: The input phase angle to apply when q2 is in the excited state and q1 is in the ground state.
:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
CPHASE = _make_gate("CPHASE", 2, 1)
"""
Produces a CPHASE00 instruction.
This gate applies to two qubit arguments to produce one of the controlled phase instructions.

:param angle: The input phase angle to apply when both qubits are in the excited state.
:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""

SWAP = _make_gate("SWAP", 2)()
"""
Produces a SWAP instruction. This gate swaps the state of two qubits.

:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
CSWAP = _make_gate("CSWAP", 3)()
"""
Produces a SWAP instruction. This gate swaps the state of two qubits.

:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
ISWAP = _make_gate("ISWAP", 2)()
"""
Produces an ISWAP instruction. This gate swaps the state of two qubits, applying a -i phase to q1 when it
is in the excited state and a -i phase to q2 when it is in the ground state.

:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""
PSWAP = _make_gate("PSWAP", 2, 1)
"""
Produces a PSWAP instruction. This is a parameterized swap gate.

:param angle: The angle of the phase to apply to the swapped states. This phase is applied to q1 when it is in
              the excited state and to q2 when it is in the ground state.
:param q1: Qubit 1.
:param q2: Qubit 2.
:returns: A Gate object.
"""

WAIT = Wait()
"""
This instruction tells the quantum computation to halt. Typically these is used while classical memory is being manipulated by a CPU in a hybrid classical/quantum algorithm.

:returns: A Wait object.
"""
RESET = Reset()
"""
This instruction resets all the qubits to the ground state.

:returns: A Reset object.
"""
NOP = Nop()
"""
This instruction applies no operation at that timestep. Typically these are ignored in error-models.

:returns: A Nop object.
"""
HALT = Halt()
"""
This instruction ends the program.

:returns: A Halt object.
"""


def MEASURE(qubit, classical_reg=None):
    """
    Produce a MEASURE instruction.

    :param qubit: The qubit to measure.
    :param classical_reg: The classical register to measure into, or None.
    :return: A Measurement instance.
    """
    qubit = unpack_qubit(qubit)
    address = None if classical_reg is None else unpack_classical_reg(classical_reg)
    return Measurement(qubit, address)


def TRUE(classical_reg):
    """
    Produce a TRUE instruction.

    :param classical_reg: A classical register to modify.
    :return: A ClassicalTrue instance.
    """
    return ClassicalTrue(unpack_classical_reg(classical_reg))


def FALSE(classical_reg):
    """
    Produce a FALSE instruction.

    :param classical_reg: A classical register to modify.
    :return: A ClassicalFalse instance.
    """
    return ClassicalFalse(unpack_classical_reg(classical_reg))


def NOT(classical_reg):
    """
    Produce a NOT instruction.

    :param classical_reg: A classical register to modify.
    :return: A ClassicalNot instance.
    """
    return ClassicalNot(unpack_classical_reg(classical_reg))


def AND(classical_reg1, classical_reg2):
    """
    Produce an AND instruction.

    :param classical_reg1: The first classical register.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalAnd instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalAnd(left, right)


def OR(classical_reg1, classical_reg2):
    """
    Produce an OR instruction.

    :param classical_reg1: The first classical register.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalOr instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalOr(left, right)


def MOVE(classical_reg1, classical_reg2):
    """
    Produce a MOVE instruction.

    :param classical_reg1: The first classical register.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalMove instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalMove(left, right)


def EXCHANGE(classical_reg1, classical_reg2):
    """
    Produce an EXCHANGE instruction.

    :param classical_reg1: The first classical register, which gets modified.
    :param classical_reg2: The second classical register, which gets modified.
    :return: A ClassicalExchange instance.
    """
    left = unpack_classical_reg(classical_reg1)
    right = unpack_classical_reg(classical_reg2)
    return ClassicalExchange(left, right)


STANDARD_GATES = {'I': I,
                  'X': X,
                  'Y': Y,
                  'Z': Z,
                  'H': H,
                  'S': S,
                  'T': T,
                  'PHASE': PHASE,
                  'RX': RX,
                  'RY': RY,
                  'RZ': RZ,
                  'CZ': CZ,
                  'CNOT': CNOT,
                  'CCNOT': CCNOT,
                  'CPHASE00': CPHASE00,
                  'CPHASE01': CPHASE01,
                  'CPHASE10': CPHASE10,
                  'CPHASE': CPHASE,
                  'SWAP': SWAP,
                  'CSWAP': CSWAP,
                  'ISWAP': ISWAP,
                  'PSWAP': PSWAP
                  }
"""
Dictionary of standard gates. Keys are gate names, values are gate functions.
"""



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from six import integer_types

from pyquil.quil import Program
from ._base_connection import validate_run_items, TYPE_MULTISHOT, TYPE_MULTISHOT_MEASURE, get_job_id, BaseConnection


class QPUConnection(BaseConnection):
    """
    Represents a connection to the QPU (Quantum Processing Unit)
    """

    def __init__(self, async_endpoint='https://job.rigetti.com/beta', api_key=None, user_id=None):
        super(QPUConnection, self).__init__(async_endpoint=async_endpoint, api_key=api_key, user_id=user_id)

    def run(self, quil_program, classical_addresses, trials=1):
        """
        Run a pyQuil program on the QPU. This functionality is in beta.

        :param Program quil_program: Quil program to run on the QPU
        :param list classical_addresses: Currently unused
        :param int trials: Number of shots to take
        :return: A list of lists of bits. Each sublist corresponds to the values
                 in `classical_addresses`.
        :rtype: list
        """
        payload = self._run_payload(quil_program, classical_addresses, trials)

        response = self._post_json(self.async_endpoint + "/job", {"machine": "QPU", "program": payload})
        job = self.wait_for_job(get_job_id(response))
        return job.result()

    def run_async(self, quil_program, classical_addresses, trials=1):
        """
        Similar to run except that it returns a job id and doesn't wait for the program to be executed.
        See https://go.rigetti.com/connections for reasons to use this method.
        """
        payload = self._run_payload(quil_program, classical_addresses, trials)
        response = self._post_json(self.async_endpoint + "/job", {"machine": "QPU", "program": payload})
        return get_job_id(response)

    def _run_payload(self, quil_program, classical_addresses, trials):
        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")
        validate_run_items(classical_addresses)
        if not isinstance(trials, integer_types):
            raise TypeError("trials must be an integer")

        payload = {"type": TYPE_MULTISHOT,
                   "addresses": classical_addresses,
                   "trials": trials,
                   "quil-instructions": quil_program.out()}

        return payload

    def run_and_measure(self, quil_program, qubits, trials=1):
        """
        Run a pyQuil program on the QPU multiple times, measuring all the qubits in the QPU
        simultaneously at the end of the program each time. This functionality is in beta.

        :param Program quil_program: A Quil program.
        :param list qubits: The list of qubits to measure
        :param int trials: Number of shots to collect.
        :return: A list of a list of bits.
        :rtype: list
        """
        payload = self._run_and_measure_payload(quil_program, qubits, trials)

        response = self._post_json(self.async_endpoint + "/job", {"machine": "QPU", "program": payload})
        job = self.wait_for_job(get_job_id(response))
        return job.result()

    def run_and_measure_async(self, quil_program, qubits, trials):
        """
        Similar to run_and_measure except that it returns a job id and doesn't wait for the program to be executed.
        See https://go.rigetti.com/connections for reasons to use this method.
        """
        payload = self._run_and_measure_payload(quil_program, qubits, trials)
        response = self._post_json(self.async_endpoint + "/job", {"machine": "QPU", "program": payload})
        return get_job_id(response)

    def _run_and_measure_payload(self, quil_program, qubits, trials):
        if not isinstance(quil_program, Program):
            raise TypeError('quil_program must be a Quil program object')
        validate_run_items(qubits)
        if not isinstance(trials, int):
            raise TypeError('trials must be an integer')

        payload = {'type': TYPE_MULTISHOT_MEASURE,
                   'qubits': qubits,
                   'trials': trials,
                   'quil-instructions': quil_program.out()}

        return payload



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from __future__ import print_function

import requests
import sys

import time
from requests.adapters import HTTPAdapter
from six import integer_types
from urllib3 import Retry

from .job import Job
from ._config import PyquilConfig

TYPE_EXPECTATION = "expectation"
TYPE_MULTISHOT = "multishot"
TYPE_MULTISHOT_MEASURE = "multishot-measure"
TYPE_WAVEFUNCTION = "wavefunction"


class BaseConnection(object):
    def __init__(self, async_endpoint=None, api_key=None, user_id=None):
        self._session = requests.Session()
        retry_adapter = HTTPAdapter(max_retries=Retry(total=3,
                                                      method_whitelist=['POST'],
                                                      status_forcelist=[502, 503, 504, 521, 523],
                                                      backoff_factor=0.2,
                                                      raise_on_status=False))

        # We need this to get binary payload for the wavefunction call.
        self._session.headers.update({"Accept": "application/octet-stream"})

        self._session.mount("http://", retry_adapter)
        self._session.mount("https://", retry_adapter)

        config = PyquilConfig()
        self.api_key = api_key if api_key else config.api_key
        self.user_id = user_id if user_id else config.user_id

        self.async_endpoint = async_endpoint

    def get_job(self, job_id):
        """
        Given a job id, return information about the status of the job

        :param str job_id: job id
        :return: Job object with the status and potentially results of the job
        :rtype: Job
        """
        response = self._get_json(self.async_endpoint + "/job/" + job_id)
        return Job(response.json())

    def wait_for_job(self, job_id, ping_time=0.1, status_time=2):
        """
        Wait for the results of a job and periodically print status

        :param job_id: Job id
        :param ping_time: How often to poll the server
        :param status_time: How often to print status, set to False to never print status
        :return: Completed Job
        """
        count = 0
        while True:
            job = self.get_job(job_id)
            if job.is_done():
                break

            if status_time and count % int(status_time / ping_time) == 0:
                if job.is_queued():
                    print("job {} is currently queued at position {}".format(job.job_id, job.position_in_queue()))
                elif job.is_running():
                    print("job {} is currently running".format(job.job_id))

            time.sleep(ping_time)
            count += 1

        return job

    def _post_json(self, url, json):
        """
        Post JSON to the Forest endpoint.

        :param str url: The full url to post to
        :param dict json: JSON.
        :return: A non-error response.
        """
        headers = {
            'X-Api-Key': self.api_key,
            'X-User-Id': self.user_id,
            'Content-Type': 'application/json; charset=utf-8'
        }
        res = self._session.post(url, json=json, headers=headers)

        # Print some nice info for unauthorized/permission errors.
        if res.status_code == 401 or res.status_code == 403:
            print("! ERROR:\n"
                  "!   There was an issue validating your forest account.\n"
                  "!   Have you run the pyquil-config-setup command yet?\n"
                  "! The server came back with the following information:\n"
                  "%s\n%s\n%s" % ("=" * 80, res.text, "=" * 80), file=sys.stderr)
            print("! If you suspect this to be a bug in pyQuil or Rigetti Forest,\n"
                  "! then please describe the problem in a GitHub issue at:\n!\n"
                  "!      https://github.com/rigetticomputing/pyquil/issues\n", file=sys.stderr)

        # Print some nice info for invalid input or internal server errors.
        if res.status_code == 400 or res.status_code >= 500:
            print("! ERROR:\n"
                  "!   Server caught an error. This could be due to a bug in the server\n"
                  "!   or a bug in your code. The server came back with the following\n"
                  "!   information:\n"
                  "%s\n%s\n%s" % ("=" * 80, res.text, "=" * 80), file=sys.stderr)
            print("! If you suspect this to be a bug in pyQuil or Rigetti Forest,\n"
                  "! then please describe the problem in a GitHub issue at:\n!\n"
                  "!      https://github.com/rigetticomputing/pyquil/issues\n", file=sys.stderr)

        res.raise_for_status()
        return res

    def _get_json(self, url):
        """
        Get JSON from a Forest endpoint.

        :param str url: The full url to fetch
        :return: Response object
        """
        headers = {
            'X-Api-Key': self.api_key,
            'X-User-Id': self.user_id,
            'Content-Type': 'application/json; charset=utf-8'
        }
        return requests.get(url, headers=headers)


def validate_noise_probabilities(noise_parameter):
    """
    Is noise_parameter a valid specification of noise probabilities for depolarizing noise?

    :param list noise_parameter: List of noise parameter values to be validated.
    """
    if not noise_parameter:
        return
    if not isinstance(noise_parameter, list):
        raise TypeError("noise_parameter must be a list")
    if any([not isinstance(value, float) for value in noise_parameter]):
        raise TypeError("noise_parameter values should all be floats")
    if len(noise_parameter) != 3:
        raise ValueError("noise_parameter lists must be of length 3")
    if sum(noise_parameter) > 1 or sum(noise_parameter) < 0:
        raise ValueError("sum of entries in noise_parameter must be between 0 and 1 (inclusive)")
    if any([value < 0 for value in noise_parameter]):
        raise ValueError("noise_parameter values should all be non-negative")


def validate_run_items(run_items):
    """
    Check the validity of classical addresses / qubits for the payload.

    :param list run_items: List of classical addresses or qubits to be validated.
    """
    if not isinstance(run_items, list):
        raise TypeError("run_items must be a list")
    if any([not isinstance(i, integer_types) for i in run_items]):
        raise TypeError("run_items list must contain integer values")


def get_job_id(response):
    return response.json()['jobId']
